
@file:Suppress("unused", "UNUSED_PARAMETER")

package ru.dbotthepony.mc.otm

import com.google.common.util.concurrent.ThreadFactoryBuilder
import net.minecraft.client.server.IntegratedServer
import net.minecraft.core.HolderLookup
import net.minecraft.server.MinecraftServer
import net.minecraft.util.RandomSource
import net.minecraft.world.level.Level
import net.neoforged.api.distmarker.Dist
import net.neoforged.fml.loading.FMLLoader
import net.neoforged.neoforge.event.server.ServerAboutToStartEvent
import net.neoforged.neoforge.event.server.ServerStoppedEvent
import net.neoforged.neoforge.event.server.ServerStoppingEvent
import net.neoforged.neoforge.event.tick.LevelTickEvent
import net.neoforged.neoforge.event.tick.ServerTickEvent
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.capability.AbstractProfiledStorage
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.core.collect.WeakHashSet
import ru.dbotthepony.mc.otm.core.util.AtomicallyInvalidatedLazy
import ru.dbotthepony.mc.otm.core.util.GJRAND64RandomSource
import ru.dbotthepony.mc.otm.core.util.IConditionalTickable
import ru.dbotthepony.mc.otm.core.util.ITickable
import ru.dbotthepony.mc.otm.core.util.TickList
import ru.dbotthepony.mc.otm.graph.GraphNodeList
import java.lang.ref.Cleaner
import java.util.*
import java.util.concurrent.atomic.AtomicInteger

private val threadLocalRandom = ThreadLocal.withInitial { GJRAND64RandomSource() }
internal val THREAD_LOCAL_RANDOM: RandomSource
	get() = threadLocalRandom.get()

private val preServerTick = TickList()
private val postServerTick = TickList()
private val preWorldTick = WeakHashMap<Level, TickList>()
private val postWorldTick = WeakHashMap<Level, TickList>()

private val clientThreads = WeakHashSet<Thread>()
private val serverThreads = WeakHashSet<Thread>()

val OTM_CLEANER: Cleaner = Cleaner.create(ThreadFactoryBuilder().setNameFormat("OverdriveThatMatters-Cleaner-%d").build())

private val serverCounter = AtomicInteger()
private var _server: MinecraftServer? = null
val isClient: Boolean by lazy { FMLLoader.getDist() == Dist.CLIENT }

val UNIVERSE_TICKS get() = postServerTick.ticks
val Level.ticksPassed get() = postWorldTick.computeIfAbsent(this) { TickList() }.ticks

@Deprecated("Use when absolutely necessary, like context-less methods/properties")
val Registries: HolderLookup.Provider get() {
	if (isClient) {
		return NULLABLE_MINECRAFT_SERVER?.registryAccess() ?: minecraft.level?.registryAccess() ?: throw IllegalStateException("Not in game")
	} else {
		return NULLABLE_MINECRAFT_SERVER?.registryAccess() ?: throw IllegalStateException("Not in game")
	}
}

fun <V> lazyPerServer(fn: (MinecraftServer) -> V): Lazy<V> {
	return AtomicallyInvalidatedLazy(serverCounter) {
		if (!SERVER_IS_LIVE)
			throw IllegalStateException("No server is running")

		fn.invoke(_server!!)
	}
}

fun onceServerPre(inTicks: Int, callback: Runnable): TickList.Timer? {
	if (!SERVER_IS_LIVE) {
		LOGGER.error("Refusing to add timer $callback in ticks $inTicks while server is dying", IllegalStateException("Server is stopping"))
		return null
	}

	return preServerTick.Timer(inTicks, callback)
}

fun onceServer(inTicks: Int, callback: Runnable): TickList.Timer? {
	if (!SERVER_IS_LIVE) {
		LOGGER.error("Refusing to add ticker $callback in ticks $inTicks while server is dying", IllegalStateException("Server is stopping"))
		return null
	}

	return postServerTick.Timer(inTicks, callback)
}

/**
 * schedules execution of Runnable somewhere in the future,
 * at discretion of tick list
 */
fun sometimeServer(callback: Runnable): TickList.Timer? {
	if (!SERVER_IS_LIVE) {
		LOGGER.error("Refusing to add ticker $callback while server is dying", IllegalStateException("Server is stopping"))
		return null
	}

	return postServerTick.sometime(callback)
}

private val isPausedImpl: Boolean get() {
	val server = _server

	if (server is IntegratedServer) {
		return !server.isPublished && minecraft.isPaused
	}

	return false
}

val isPaused: Boolean get() {
	if (clientThreads.isEmpty()) {
		return false
	}

	return isPausedImpl
}

internal fun recordClientThread() {
	clientThreads.add(Thread.currentThread())
}

fun runIfClient(lambda: () -> Unit) {
	try {
		if (isClient) {
			lambda.invoke()
		}
	} catch(err: RuntimeException) {
		if (err.message?.startsWith("Attempted to load class") != true) {
			throw err
		}

		LOGGER.error("Unexpected classloading exception", err)
	}
}

fun <V> runIfClient(value: V, lambda: () -> V): V {
	try {
		if (isClient) {
			return lambda.invoke()
		}

		return value
	} catch(err: RuntimeException) {
		if (err.message?.startsWith("Attempted to load class") == true) {
			LOGGER.error("Unexpected classloading exception", err)
			return value
		}

		throw err
	}
}

fun <V> runOnClient(value: V, lambda: () -> V): V {
	if (isClientThread()) {
		return lambda.invoke()
	}

	return value
}

fun isServerThread(): Boolean {
	return Thread.currentThread() in serverThreads
}

fun isClientThread(): Boolean {
	return Thread.currentThread() in clientThreads
}

val MINECRAFT_SERVER: MinecraftServer
	get() {
		return _server ?: throw NullPointerException("Not running a server!")
	}

val NULLABLE_MINECRAFT_SERVER: MinecraftServer?
	get() {
		return _server
	}

// Flag whenever is server alive
// To avoid baby deadlocks caused by minecraft engine design issues
var SERVER_IS_LIVE = false
	private set

private val LOGGER = LogManager.getLogger()

internal fun onServerTickPre(event: ServerTickEvent.Pre) {
	preServerTick.tick()
	serverThreads.add(Thread.currentThread())
}

internal fun onServerTickPost(event: ServerTickEvent.Post) {
	postServerTick.tick()
	// чтоб не плодить кучу подписчиков, вызовем напрямую отсюда
	GraphNodeList.tick()
	AbstractProfiledStorage.onServerPostTick()
}

internal fun onLevelTickPre(event: LevelTickEvent.Pre) {
	preWorldTick[event.level]?.tick()

	if (event.level.isClientSide) {
		clientThreads.add(Thread.currentThread())
	} else {
		serverThreads.add(Thread.currentThread())
	}
}

internal fun onLevelTickPost(event: LevelTickEvent.Post) {
	postWorldTick[event.level]?.tick()
}

fun onceServerPre(ticker: ITickable) {
	preServerTick.once(ticker, SERVER_IS_LIVE, "Server is stopping")
}

fun onceServer(ticker: ITickable) {
	postServerTick.once(ticker, SERVER_IS_LIVE, "Server is stopping")
}

fun tickServerPre(ticker: IConditionalTickable) {
	preServerTick.add(ticker, SERVER_IS_LIVE, "Server is stopping")
}

fun tickServer(ticker: IConditionalTickable) {
	postServerTick.add(ticker, SERVER_IS_LIVE, "Server is stopping")
}

fun Level.once(time: Int, ticker: Runnable): TickList.Timer? {
	if (this.isClientSide) return null

	if (!SERVER_IS_LIVE) {
		LOGGER.error("Refusing to add ticker $ticker while server is dying", IllegalStateException("Server is stopping"))
		return null
	}

	return postWorldTick.computeIfAbsent(this) { TickList() }.timer(time, ticker)
}

fun Level.once(ticker: ITickable) {
	if (this.isClientSide) return

	if (!SERVER_IS_LIVE) {
		LOGGER.error("Refusing to add ticker $ticker while server is dying", IllegalStateException("Server is stopping"))
		return
	}

	postWorldTick.computeIfAbsent(this) { TickList() }.once(ticker)
}

fun Level.oncePre(ticker: ITickable) {
	if (this.isClientSide) return

	if (!SERVER_IS_LIVE) {
		LOGGER.error("Refusing to add ticker $ticker while server is dying", IllegalStateException("Server is stopping"))
		return
	}

	preWorldTick.computeIfAbsent(this) { TickList() }.once(ticker)
}

fun Level.addTicker(ticker: IConditionalTickable) {
	if (this.isClientSide) return

	if (!SERVER_IS_LIVE) {
		LOGGER.error("Refusing to add ticker $ticker while server is dying", IllegalStateException("Server is stopping"))
		return
	}

	postWorldTick.computeIfAbsent(this) { TickList() }.add(ticker)
}

fun Level.addTickerPre(ticker: IConditionalTickable) {
	if (this.isClientSide) return

	if (!SERVER_IS_LIVE) {
		LOGGER.error("Refusing to add ticker $ticker while server is dying", IllegalStateException("Server is stopping"))
		return
	}

	preWorldTick.computeIfAbsent(this) { TickList() }.add(ticker)
}

fun Level.until(ticker: () -> Boolean) {
	if (this.isClientSide) return
	addTicker(IConditionalTickable.wrap(ticker))
}

fun Level.untilPre(ticker: () -> Boolean) {
	if (this.isClientSide) return
	addTickerPre(IConditionalTickable.wrap(ticker))
}

fun Level.`while`(condition: () -> Boolean, ticker: () -> Unit) {
	if (this.isClientSide) return
	addTicker(IConditionalTickable.wrap(condition, ticker))
}

fun Level.whilePre(condition: () -> Boolean, ticker: () -> Unit) {
	if (this.isClientSide) return
	addTickerPre(IConditionalTickable.wrap(condition, ticker))
}

private fun clear() {
	preServerTick.clear()
	postServerTick.clear()
	preWorldTick.clear()
	postWorldTick.clear()
}

internal fun onServerStarting(event: ServerAboutToStartEvent) {
	clear()
	SERVER_IS_LIVE = true
	_server = event.server
	serverThreads.add(Thread.currentThread())
	serverCounter.incrementAndGet()
}

internal fun onServerStopping(event: ServerStoppingEvent) {
	clear()
	SERVER_IS_LIVE = false
	serverCounter.incrementAndGet()
}

internal fun onServerStopped(event: ServerStoppedEvent) {
	if (SERVER_IS_LIVE) {
		LOGGER.fatal("ServerStoppingEvent did not fire. If server has crashed this is normal. However, if server finished it's work 'gracefully' this is a bug!")

		clear()
		SERVER_IS_LIVE = false
	}

	_server = null
	serverCounter.incrementAndGet()
}
